/*                                      */
/* Out(F_n) computation                 */
/* Ferenc Gerlits, PhD thesis, Aug 2002 */
/*                                      */
/* Permutation class                                           */
/* with a 'next' function which steps through all permutations */
/*                                                             */

package OutH;

public class Permutation extends Object 
                         implements java.io.Serializable
{
  public int size;
  public int p[];
  
  public Permutation(int s) {
    int i;
    
    size = s;
    p = new int[size];
    for (i=0; i<size; i++)
      p[i] = i;
    }
    
  public Permutation(Permutation sigma) {
    int i;
    
    size = sigma.size;
    p = new int[size];
    for (i=0; i<size; i++)
      p[i] = sigma.p[i];
    }

    
  public Permutation(int s, int[] b) {
    int i;
    
    size = s;
    p = new int[size];
    for (i=0; i<size; i++)
      p[i] = b[i];
    }
    
    
  public int sign() {
    int i, j, inversions = 0;
    
    for (i=0; i<size; i++)
      for (j=i+1; j<size; j++)
        if (p[j] < p[i])
          inversions++;
    if (inversions % 2 == 0)
      return 1;
    else return -1;
    }
    
    
  public boolean next() {
    int i, j, first;
    int q[];
    Permutation sigma;
    
    if (size == 0 || size == 1)
      return false;
      
    if (size == 2) {
      if (p[0] == 0) {
        p[0] = 1;
        p[1] = 0;
        return true;
        }
      else {
        p[0] = 0;
        p[1] = 1;
        return false;
        }
      }
    
    first = p[0];					// cut off first number,
    q = new int[size-1];			// e.g., (2314) -> (213)
    for (i=1; i<size; i++) {
      if (p[i] < first)
        q[i-1] = p[i];
      else
        q[i-1] = p[i] - 1;
      }					

    sigma = new Permutation(size-1, q);
    
    if (sigma.next()) {				// advance 'tail', if possible
	  for (i=1; i<size; i++) {
	    if (sigma.p[i-1] < first)
	      p[i] = sigma.p[i-1];
	    else
	      p[i] = sigma.p[i-1] + 1;
	    }
	  }
	else {
	  if (first+1 < size) {			// new first element, the
	    p[0] = ++first;				// rest are in increasing
	    for (i=0; i<size; i++) {	// order
	      if (i < first)
	        p[i+1] = i;
	      if (i > first)
	        p[i] = i;
	      }
	    }
	  else {
	    for (i=0; i<size; i++)		// return 'next' permutation;
	      p[i] = i;					// if we were at the end, 'next'
	    return false;				// is back to the first
	    }
	  }
	  
	return true;
	}
	
	
  public static Permutation multiply(Permutation s1, Permutation s2) {
    int i;
    Permutation sigma;
    
    if (s1.size == s2.size)
      sigma = new Permutation(s1); 	
    else return null;
    
    sigma.rightMultiply(s2);
    
    return sigma;
    }


  public void leftMultiply(Permutation sigma) {
    int i, b[] = new int[size];
    
    if (sigma.size != size)		// shouldn't happen; make nicer later
      return;
    
    for (i=0; i<size; i++)		// apply sigma first, then this
      b[i] = this.p[sigma.p[i]];

    for (i=0; i<size; i++)
      p[i] = b[i];
    
    }
    

  public void rightMultiply(Permutation sigma) {
    int i, b[] = new int[size];
    
    if (sigma.size != size)		// shouldn't happen; make nicer later
      return;
    
    for (i=0; i<size; i++)		// apply this first, then sigma
      b[i] = sigma.p[this.p[i]];

    for (i=0; i<size; i++)
      p[i] = b[i];
    
    }


  public void conjugate(Permutation t) {
    leftMultiply(t.inverse());
    rightMultiply(t);
    }


  public Permutation inverse() {
    int i;
    Permutation sigma = new Permutation(size);
    
    for (i=0; i<size; i++)
      sigma.p[this.p[i]] = i;
    return sigma;
    }
    
  
  public static Permutation concatenate(Permutation s1, Permutation s2) {
    int i;
    Permutation sigma = new Permutation(s1.size + s2.size);
    
    for (i=0; i<s1.size; i++)
      sigma.p[i] = s1.p[i];
      
    for (i=0; i<s2.size; i++)
      sigma.p[s1.size + i] = s2.p[i] + s1.size;
      
    return sigma;
    }
  
  
  public static Permutation concatenate(Permutation s1, Permutation s2,
                                              Permutation s3) {
    int i;
    Permutation sigma = new Permutation(s1.size + s2.size + s3.size);
    
    for (i=0; i<s1.size; i++)
      sigma.p[i] = s1.p[i];
      
    for (i=0; i<s2.size; i++)
      sigma.p[s1.size + i] = s2.p[i] + s1.size;

    for (i=0; i<s3.size; i++)
      sigma.p[s1.size + s2.size + i] = s3.p[i] + s1.size + s2.size;
      
    return sigma;
    }
  
  
  public String toString() {
    int i;
    String str = new String("(");
    
    for (i=0; i<size-1; i++) {
      str = str.concat(String.valueOf(p[i] + 1));
      str = str.concat(",");
      }
    str = str.concat(String.valueOf(p[size-1] + 1));
    str = str.concat(")");
    return str;
    }
    
}
