/*                                      */
/* Out(F_n) computation                 */
/* Ferenc Gerlits, PhD thesis, Aug 2002 */
/*                                      */
/* SimpleGraph class; contains a matrix representation of a graph */
/* has two children: Graph and Tree                               */
/*                                                                */

package OutH;

import java.util.Vector;

public class SimpleGraph extends Object implements java.io.Serializable
{
  public int size;			// number of vertices
  public byte m[][];		// to store the matrix 
  public int noEdges;		// # of edges (double edges are 1, loops 0)
  public int allEdges;		// # of all edges, with loops and all
  public int noLoops;		// # of loops; these always come first in NF
  public int noMults;		// # of vertices with multiple edges; these 
                            //  are next, before the rest of the vert's
  
  public SimpleGraph(int n) {
    size = n;
    m = new byte[size][size];
    noEdges = allEdges = noLoops = noMults = 0;
    }
   
   
  public SimpleGraph(SimpleGraph g) {
    this(g.size);
    int i,j;
    
    for (i=0; i<size; i++)		// only storing #s on or above
      for (j=i; j<size; j++)	//   the diagonal
        m[i][j] = g.m[i][j];
    noEdges = g.noEdges;
    allEdges = g.allEdges;
    noLoops = g.noLoops;
    noMults = g.noMults;
    }
    

  public void permute(Permutation sigma) {
    int i, j, x, y, t;						// sigma is a permutation
    byte b[][] = new byte[size][size];		//   of 'size' numbers
    								
    for (i=0; i<size; i++)
      for (j=i; j<size; j++) {
        x = sigma.p[i];			
        y = sigma.p[j];			
        if (x > y) 			
          { t=x; x=y; y=t; }
        b[x][y] = m[i][j];
        }	
    for (i=0; i<size; i++)
      for (j=i; j<size; j++)
        m[i][j] = b[i][j];			

    }
    
    
  public boolean equals(SimpleGraph otherg) {
    int i, j;
    
    if (size != otherg.size ||
        noEdges != otherg.noEdges ||
        allEdges != otherg.allEdges ||
        noLoops != otherg.noLoops ||
        noMults != otherg.noMults)
      return false;
    
    for (i=0; i<size; i++)
      for (j=i; j<size; j++)
        if (m[i][j] != otherg.m[i][j])
          return false;
    return true;
    }
    
    
  public static int order(SimpleGraph g1, SimpleGraph g2) {
    int i, j;			
    								// return 0 if the graphs are equal,
    for (i=0; i<g1.size; i++)		// -1 if g1<g2 and 1 if g1>g2
      for (j=i; j<g1.size; j++) {
        if (g1.m[i][j] < g2.m[i][j])	// not checking size; this 
          return -1;					// should only be called
        if (g1.m[i][j] > g2.m[i][j])	// when they're equal
          return 1;
        }
    return 0;
    }


  public Permutation toNormalForm(Vector symBag) {
    int i, j;					// put graph in normal form
    SimpleGraph g, gmax;		// and find its automorphisms
    Permutation sigma, sigma1, sigma2, sigma3, sstored, tau;
    
    Vector tempBag = new Vector(10,10);
    sigma1 = new Permutation(noLoops);
    sigma2 = new Permutation(noMults);
    sigma3 = new Permutation(size - noLoops - noMults);
    sstored = new Permutation(size);
    gmax = new SimpleGraph(this);

    do {
      do {
        do {
          sigma = Permutation.concatenate(sigma1, sigma2, sigma3);
          g = new SimpleGraph(this);
          g.permute(sigma);
          switch (order(g, gmax)) {
            case 0:  tempBag.addElement(new Permutation(sigma));
                     break;			// new automorphism
            case 1:  gmax = g;							// new lexico-
                     sstored = new Permutation(sigma);	// graphic max
                     tempBag.removeAllElements();
                     tempBag.addElement(sigma);
                     break;
            }
          } while (sigma3.next());
        } while (sigma2.next());
      } while (sigma1.next());
    
    for (i=0; i<size; i++)		
      for (j=i; j<size; j++)
        m[i][j] = gmax.m[i][j];
    
    tau = sstored.inverse();
    for (i=0; i<tempBag.size(); i++) {
      sigma = (Permutation)tempBag.elementAt(i);
      sigma.leftMultiply(tau);
      symBag.addElement(sigma);
      }

    return sstored;
    }

  
  public int[] kthEdge(int k) {		// find kth edge of the graph
    int i, j, n, 
        r[] = new int[2];
        
    n = -1;
    for (i=0; i<size; i++)
      for (j=i; j<size; j++) {
        n += m[i][j];
        if (n >= k) {
          r[0] = i;
          r[1] = j;
          return r;
          }
        }
    r[0] = r[1] = -1;				// error: there is no kth edge
    return r;
    }


  public void contractEdge(int k) {
    int r[];						// find the index of the
    								// kth edge, and call
    r = this.kthEdge(k);			// the version below    
    contractEdge(r[0], r[1]);
    }
  

  public void contractEdge(int x, int y) {
    int i, j;
    
    if (x >= y)	{		// can't contract a loop!
      System.out.println("Attempt to contract a loop in SimpleGraph.contractEdge.");
      System.exit(1);
      }
      
    byte b[][] = new byte[size - 1][size - 1];
        
    for (i=0; i<x; i++) {			// rows 1 through x-1
      for (j=i; j<x; j++)
        b[i][j] = m[i][j];
      b[i][x] = (byte)(m[i][x] + m[i][y]);
      for (j=x+1; j<y; j++)
        b[i][j] = m[i][j];
      for (j=y+1; j<size; j++)
        b[i][j-1] = m[i][j];
      }
    
    b[x][x] = (byte)(m[x][x] + m[y][y] + m[x][y] - 1);	// row x
    for (j=x+1; j<y; j++)
      b[x][j] = (byte)(m[x][j] + m[j][y]);
    for (j=y+1; j<size; j++)
      b[x][j-1] = (byte)(m[x][j] + m[y][j]);
      
    for (i=x+1; i<y; i++) {			// rows x+1 through y-1
      for (j=i; j<y; j++)
        b[i][j] = m[i][j];
      for  (j=y+1; j<size; j++)
        b[i][j-1] = m[i][j];
      }
      
    for (i=y+1; i<size; i++)		// rows y+1 through 'size'
      for (j=i; j<size; j++)
        b[i-1][j-1] = m[i][j];
        
    size--;
    m = b;    
    allEdges--;
    // g.noEdges, g.noLoops, g.noMults need to be fixed;
    // this is done differently in classes Graph and Tree
    }  
    
  
  public String toString() {
    int i, j;
    String str = new String();
    
    for (i=0; i<size; i++) {
      str = str + "[";
      for (j=0; j<i; j++)
        str = str + "\t";
      for (j=i; j<size; j++) { 
        str = str + "\t" + m[i][j];
        }
      str = str + "\t]\r\n";
      }
      
    str = str +"% "+ noEdges +", "+ allEdges +", "+ noLoops +", "+ noMults;
    return str;
    }  
  
/*  
  public static void main(String args[]) {		// testing...
    byte sample[][] = {{1, 1, 1, 1, 1 },
                       {0, 2, 2, 1, 1 },
                       {0, 0, 3, 1, 1 },
                       {0, 0, 0, 1, 1 },
                       {0, 0, 0, 0, 0 }};
    SimpleGraph g = new SimpleGraph(5);
    
    for (int i=0; i<g.size; i++)
      for (int j=i; j<g.size; j++)
        g.m[i][j] = sample[i][j];
    g.noEdges = 7;
    g.allEdges = 18;
    g.noLoops = 4;
    g.noMults = 0;
    
    System.out.println(g);
    
    Vector symBag = new Vector(10,10);
    Permutation sigma = g.toNormalForm(symBag);
    
    System.out.println(sigma);
    System.out.println(g);
    System.out.print(symBag.size()+": ");
    for (int i=0; i<symBag.size(); i++)
      System.out.print((Permutation)symBag.elementAt(i)+", ");
    }
*/    
}
